'''This code is released under a Creative Commons Attribution-NoDerivs
3.0 Unported License (http://creativecommons.org/licenses/by-nd/3.0/).
If you want to do something with the code, email me and I'll grant you
permission (in all likelihood)- I'd just like to know what people are
up to.

-Matt Keeter
mkeeter@hmc.edu
mattkeeter.com

This code comes with no warranty whatsoever.
'''
currentPlaylist = "5 Stars"

from commands import getoutput
from os import system, getcwd
from subprocess import Popen, PIPE
from sys import argv

from math import log10, floor
from random import choice

import curses
from curses import KEY_LEFT, KEY_RIGHT

STATUS_PATH = getcwd() + "/" + argv[0]
if "//" in STATUS_PATH: # fixes bug in invocation
    STATUS_PATH = argv[0]

################################################################################
# Apple-script related functions
################################################################################
def formatCommand(script):
    command = "osascript"
    while script.find("\t") != -1: script = script.replace("\t","")
    while script != '':
        if script.find("\n") != -1:
            nextPart = script[0:script.index("\n")]
            script = script[script.index("\n")+1:]
        else:
            nextPart = script
            script = ""
        if nextPart != '': command += " -e '" + nextPart + "'"   
    return command

def getScriptOutput(script):
    command = formatCommand(script)
    return getoutput(command)

def runScript(script):
    command = formatCommand(script)
    system(command)

################################################################################
# Status-related helper functions
################################################################################
currentdataScript = """
tell application "iTunes"
    {name of current track, "@@@", artist of current track, "@@@", \
    album of current track, "@@@", duration of current track, "@@@", \
    player position, "@@@", player state}
end tell
"""

# checks if this is a status subprocess
if "status" in argv:
    if system("ps -ea|grep iTunes|grep -v grep|grep -v Helper") == 0:
        print "None, @@@, None, @@@, None, @@@, 0.0, @@@, 0, @@@, Off"
        quit()
    print getScriptOutput(currentdataScript)
    quit()

def doStatus(currentData):
    maxdraw = int(width/2 - 3)
    splitData = currentData.split(", @@@, ")
    if len(splitData) == 1:
        splitData = ["None","None","None","0.0", "0","stopped"]
    status.clear()
    drawSeparators(maxdraw)
    printCurrentSong(splitData, maxdraw)
    printSongState(splitData)
    status.refresh()

def drawSeparators(maxdraw):
    for i in range(width): status.addch(0,i,"_") # top separator
    for i in range(1,4): status.addch(i, maxdraw+1, "|")

def drawProgressLine( position, length ):
    bar = "["+"-"*int(width/2-3)+"]"
    status.addstr(1, int(width/2), bar)
    if float(length) != 0:
        status.addch(1, int( (width/2 - 3) * float(position)/float(length)) + int(width/2 + 1), "|")

def sec_to_time(seconds):
    mins = int(seconds)/60
    secs = int(seconds) - mins*60
    return '0'*(2-len(str(mins)))+str(mins)+":"+'0'*(2-len(str(secs)))+str(secs)

def printCurrentSong(data, maxdraw):
    status.addnstr(1, 0, data[0], maxdraw)
    status.addnstr(2, 0, data[1], maxdraw)
    status.addnstr(3, 0, data[2], maxdraw)

def printSongState(data):
    drawProgressLine(data[4], data[3]) # progress bar
    status.addstr(2, width-19, data[5]) # current iTunes state
    status.addstr(2,width-11,sec_to_time(data[4])) # current time
    status.addstr(2, width-6, "/"+sec_to_time(data[3].split(".")[0]) ) # total time
   

################################################################################
# Command prompt related helper functions
################################################################################
helpText = '''
p: toggle playing/paused state
pr: play random song
n / next / ->: go to next song
b / back / <-: go to previous song
r: shows random songs
s NAME: searches for songs by name
sa ARTIST: searches for songs by artist
clear: clear screen
quit: quit program
'''

psongScript = """
tell application "iTunes"
    play (first track of playlist "PLIST" where name is "NAME")
end tell
"""

commandDictionary = \
{ "help": "printResponse(helpText)", \
  "clear": "clearConsole()", \
  "next": """runScript('tell application "iTunes" to next track'); newLine()""", \
  "back": """runScript('tell application "iTunes" to back track'); newLine()""", \
  "p": """runScript('tell application "iTunes" to playpause'); newLine()""", \
  "pr":"""playRandom(); newLine()""",\
  "quit": """global quitter; quitter = True""", \
  "r":"""printResponse(map(formatSong, [choice(songList) for x in range(15)]))""", \
  "s": """search('''ARG'''.lower())""", \
  "sa":"""searchArtist('''ARG'''.lower())""",\
  "psong":"""playSong('''ARG'''); newLine()""" }

commandSets = [\
    ["next","n"],\
    ["back","b"],\
    ["quit","exit","q"]]

def doCommand(fullCommand):
    splitCommand = fullCommand.split(" ")
    command = splitCommand[0]
    if len(splitCommand) > 1:
        arg = fullCommand[fullCommand.index(" ")+1:]
        if command == "p": command = "psong"
    else:
        arg = ""
    for cSet in commandSets: # deal with multiple commands doing the same thing
        if command in cSet: command = cSet[0]
    if command in commandDictionary.keys():
        exec( commandDictionary[command].replace("ARG", arg) )
    else:
        printResponse('"'+command+'": Command not found.')

def search(songName):
    if songName == "":
        printResponse('''Usage: "s SONGNAME"''')
        return 2
    songArray = []
    for song in songList:
        if songName in song["name"].lower(): songArray += [song]
    if songArray != []:
        printResponse( map(formatSong, songArray) )
        return 1
    else:
        printResponse( "No results found." )
        return 0

def searchArtist(artist):
    if artist == "":
        printResponse('''Usage: "sa ARTIST"''')
        return 2
    songArray = []
    for song in songList:
        if artist in song["artist"].lower(): songArray += [song]
    if songArray != []:
        printResponse( map(formatSong, songArray) )
        return 1
    else:
        printResponse( "No results found." )
        return 0

def playSong(songName):
    if songName == "":
        printResponse('''Usage: "p SONGNAME"''')
        return 2
    songArray = []
    bestMatch = [{}, 0]
    for song in filter(lambda(x): songName.lower() in x["name"].lower(), songList):
        percentageMatch = float(len(songName))/float(len(song["name"]))
        if  percentageMatch > bestMatch[1]:
            bestMatch[0] = song
            bestMatch[1] = percentageMatch        
    if bestMatch[0] == {}:
        printResponse( "No results found." )
        return 0
    else:
        c = psongScript.replace("NAME",bestMatch[0]["name"])
        c = c.replace("PLIST",currentPlaylist)
        runScript(c)
        return 0

def playRandom():
    songName = choice(songList)["name"]
    c = psongScript.replace("NAME",songName)
    c = c.replace("PLIST",currentPlaylist)
    runScript(c)

def clearConsole():
    global lineNum
    console.move(0,0)
    for line in range(lineNum+1):
        console.deleteln()
    lineNum = 0

def newLine():
    global lineNum
    if lineNum > height - 6:
        console.move(0,0)
        console.deleteln()
    else:
        lineNum += 1
    
def formatSong(song):
    full_name = song["name"]
    artist = song["artist"]
    album = song["album"]
    section_size = int(width / 3)
    if len(full_name) > section_size - 4:
            full_name = full_name[0:section_size-4]+' '*4
    while len(full_name) < section_size:
            full_name = full_name + ' '
    if len(artist) > section_size - 4:
            artist = artist[0:section_size-4]+' '*4
    while len(artist) < section_size:
            artist = artist + ' '
    if len(album) > section_size-4:
            album = album[0:section_size]#+' '*4
    while len(album) < section_size:
            album = album + ' '
    return full_name + artist + album
    
def printResponse(data):
    global lineNum
    if type(data) == str:
        response = data.split("\n")
    elif type(data) == list:
        response = data
        
    newLine()
        
    for line in response:
        if type(line) == str:
            console.addnstr(lineNum, 0, line, width)
        newLine()
            
################################################################################
# refresh-related scripts
################################################################################
refreshScript = """
tell application "iTunes"
	set plistName to "PLIST"
	set namelist to (name of every track in playlist plistName)
	set artistlist to (artist of every track in playlist plistName)
	set albumlist to (album of every track in playlist plistName)
	set parsedlist to {}
	set i to 1
	repeat while i is less than length of namelist
		set end of parsedlist to (item i of namelist & "###" & item i of artistlist & "###" & item i of albumlist & "@@@")
		set i to i + 1
	end repeat
	parsedlist
end tell
"""

def refresh():
    unparsedString = getScriptOutput(refreshScript.replace("PLIST",currentPlaylist))
    global songList
    for songStuff in unparsedString.split("@@@"):
        splitSong = songStuff.split("###")
        if splitSong == ['']: break
        if splitSong[0][0] == ",":
            name = splitSong[0][2:]
        else:
            name = splitSong[0]
        songList += [{"name":name, "artist":splitSong[1], "album":splitSong[2]}]


################################################################################
# main
################################################################################
def main(screen):
    console.addnstr(0,0,"tunes 2.0",width)
    
    console.addnstr(2,0,"Loading songs...",width)
    console.refresh()
    refresh()

    console.addnstr(3,0,"Updating status...",width)
    console.refresh()
    status = getoutput("python "+STATUS_PATH+" status")
    
    console.clear()
    
    commandBuffer = ""
    global lineNum

    console.addstr(0,0,"> ")
    doStatus(status)
    
    status = Popen(["python", STATUS_PATH, "status"], stdout = PIPE)
    
    console.refresh()
    while True:
        if status.poll() != None:   # check if status has updated
            doStatus(status.communicate()[0])
            status = Popen(["python", STATUS_PATH, "status"], stdout = PIPE)
            
        c = console.getch()
        if c == 10: # enter
            doCommand(commandBuffer)
            commandBuffer = ""
        elif c == 127: # backspace
            commandBuffer = commandBuffer[:-1]
            console.addch(lineNum, len(commandBuffer)+2, " ")
        elif c == 32 and commandBuffer == "":   # space bar == playpause
            runScript('tell application "iTunes"to playpause')
        elif c == KEY_RIGHT:                    # right arrow == next
            runScript('tell application "iTunes" to next track')
        elif c == KEY_LEFT:                     # left arrow == back
            runScript('tell application "iTunes" to back track')
        elif c >= 32 and c <= 126:
            commandBuffer += chr(c)

        if quitter:         # if we want to quit...
            status.wait()   # wait for subprocess to terminate
            curses.echo()
            curses.nocbreak()
            break           # then die
            
        console.addnstr(lineNum, 0, "> "+commandBuffer, width)
        console.refresh()

################################################################################
# and here we go!
################################################################################

# magic starts here
import termios, fcntl, struct, sys
s = struct.pack("HHHH", 0, 0, 0, 0)
fd_stdout = sys.stdout.fileno()
x = fcntl.ioctl(fd_stdout, termios.TIOCGWINSZ, s)
sizes=struct.unpack("HHHH", x)
width=sizes[1]
height=sizes[0]
# magic ends here
# magic is from http://bytes.com/forum/thread607757.html

# general globals
songList = []
lineNum = 0
quitter = False

# now we set up curses
stdscr = curses.initscr()
curses.start_color()
curses.meta(True)
curses.use_default_colors()
console = curses.newwin(height-4, width, 0, 0)
status = curses.newwin(4, width, height-4, 0)

console.keypad(True)
console.nodelay(True)

# start
curses.wrapper(main)
